package mnet

  1. Overview
  2. Docs

v0.0.2 (2026-05-06)

  • Improve our unikernel example and our resolver, it's able to take a domain-name as an argument (@hannesm, !14)
  • Improve our IPv4 routing (@hannes, @reynir, @dinosaure, !16)
  • Improve mnet-cli and add titles for some options (@dinosaure, !17)
  • Catch exceptions when we read on TCP and/or TLS connections for DNS packets (@dinosaure, !18)
  • Fix how we try to instantiate a connection to nameservers (@dinosaure, @hannesm, !19)
  • Fix how we push new happy-eyeballs actions (@dinosaure, !20)
  • Add options for happy-eyeballs (@dinosaure, !22)
  • Fix our mnet-tls implementation and pass a flow after its handshake even if the flow was closed (on read and write side) (@dinosaure, !23)
  • Try to send TCP packets from our TCP timer directly without interruptions instead of push it into our queue (@dinosaure, !24)
  • Be able to resolve domain-name into our happy-eyeballs implementation (@dinosaure, !25)
  • Fix our mnet-dns implementation:

    • we fix how we read DNS packet from TCP and/or TLS connections
    • we fix how we try to instantiate a new connection to nameservers

    (@dinosaure, !26)

v0.0.1 (2026-02-17)

  • First public release of mnet